package eu.nk2.cache

import org.slf4j.LoggerFactory
import org.springframework.cache.Cache
import org.springframework.cache.CacheManager
import reactor.core.publisher.Flux
import reactor.core.publisher.Mono
import reactor.core.publisher.Signal
import java.util.concurrent.ConcurrentHashMap

abstract class CachedService {
  val log = LoggerFactory.getLogger(javaClass)

  abstract val cacheTag: String
  abstract val manager: CacheManager

  val defaultCacheStore by lazy {
    manager.getCache(cacheTag)
      ?: error("cache $cacheTag cannot be created")
  }

  private val customCacheStores: ConcurrentHashMap<String, Cache> = ConcurrentHashMap()

  fun buildCacheId(id: String, vararg objects: Any): String =
    "$id${if(objects.isNotEmpty()) ":" else ""}" + objects.asSequence()
      .map { it.toString() }
      .sorted()
      .joinToString(separator=",") { it }


  inline fun <reified V> monoCached(id: Any, storePrefix: String? = null, noinline supplier: () -> Mono<V>): Mono<V> =
    CacheMono.lookup(monoReader<V>(getCustomCacheStoreByPrefix(storePrefix)), id)
      .onCacheMissResume(supplier)
      .andWriteWith(monoWriter<V>(getCustomCacheStoreByPrefix(storePrefix)))
      .onErrorEvict(monoEvicter(getCustomCacheStoreByPrefix(storePrefix)))

  inline fun <reified V> fluxCached(id: Any, storePrefix: String? = null, noinline supplier: () -> Flux<V>): Flux<V> =
    CacheFlux.lookup(fluxReader<V>(getCustomCacheStoreByPrefix(storePrefix)), id)
      .onCacheMissResume(supplier)
      .andWriteWith(fluxWriter<V>(getCustomCacheStoreByPrefix(storePrefix)))
      .onErrorEvict(fluxEvicter(getCustomCacheStoreByPrefix(storePrefix)))

  fun <T> Flux<T>.withEvictedCache(ids: List<Any>, storePrefix: String? = null) =
    this.collectList()
      .flatMap { cacheReset(getCustomCacheStoreByPrefix(storePrefix), ids.toTypedArray(), it) }
      .flatMapIterable { it }

  fun <T> Flux<T>.withEvictedCache(vararg ids: Any, storePrefix: String? = null) =
    this.collectList()
      .flatMap { cacheReset(getCustomCacheStoreByPrefix(storePrefix), ids, it) }
      .flatMapIterable { it }

  fun <T> Flux<T>.withEvictedStore(storePrefix: List<String>) =
    this.collectList()
      .flatMap { cacheClearReset(storePrefix.map { getCustomCacheStoreByPrefix(it) }, supplierValue = it) }
      .flatMapIterable { it }

  fun <T> Flux<T>.withEvictedStore(vararg storePrefix: String) =
    this.collectList()
      .flatMap { cacheClearReset(storePrefix.map { getCustomCacheStoreByPrefix(it) }, supplierValue = it) }
      .flatMapIterable { it }

  fun <T> Mono<T>.withEvictedCache(ids: List<Any>, storePrefix: String? = null) =
    this.flatMap { cacheReset(getCustomCacheStoreByPrefix(storePrefix), ids.toTypedArray(), it) }

  fun <T> Mono<T>.withEvictedCache(vararg ids: Any, storePrefix: String? = null) =
    this.flatMap { cacheReset(getCustomCacheStoreByPrefix(storePrefix), ids, it) }

  fun <T> Mono<T>.withEvictedStore(storePrefix: List<String>) =
    this.flatMap { cacheClearReset(storePrefix.map { getCustomCacheStoreByPrefix(it) }, supplierValue = it) }

  fun <T> Mono<T>.withEvictedStore(vararg storePrefix: String) =
    this.flatMap { cacheClearReset(storePrefix.map { getCustomCacheStoreByPrefix(it) }, supplierValue = it) }

  fun getCustomCacheStoreByPrefix(storePrefix: String?) =
    if(storePrefix != null) customCacheStores[storePrefix]
      ?: manager.getCache("$cacheTag::$storePrefix")
        ?.apply { customCacheStores[storePrefix] = this }
      ?: error("cache $cacheTag::$storePrefix cannot be created")
    else null

  private fun <T> cacheClearReset(cacheStore: List<Cache?>, supplierValue: T): Mono<T> =
    Mono.fromCallable {
      log.debug("[$cacheTag] clear-evicting cache ${cacheStore.joinToString(separator = ",") { it?.name ?: "null" }}")
      cacheStore.forEach { (it ?: defaultCacheStore).clear() }
      supplierValue
    }

  private fun <T> cacheReset(cacheStore: Cache?, ids: Array<out Any>, supplierValue: T): Mono<T> =
    Mono.fromCallable {
      log.debug("[$cacheTag] evicting cache ${ids.joinToString(separator=",") { it.toString() }}")
      ids.forEach { (cacheStore ?: defaultCacheStore).evict(it) }
      supplierValue
    }

  inline fun <reified V> monoReader(cacheStore: Cache?): CacheMono.MonoCacheReader<Any, V> =
    CacheMono.MonoCacheReader<Any, V> { key ->
      log.debug("[${cacheStore?.name ?: cacheTag}] reading cache $key")
      Mono.justOrEmpty<V>((cacheStore ?: defaultCacheStore).get(key, V::class.java))
        .map { Signal.next(it) }
    }

  inline fun <reified V> monoWriter(cacheStore: Cache?): CacheMono.MonoCacheWriter<Any, V> =
    CacheMono.MonoCacheWriter<Any, V> { key, value: Signal<V> ->
      log.debug("[${cacheStore?.name ?: cacheTag}] writing cache $key")
      Mono.fromRunnable { (cacheStore ?: defaultCacheStore).put(key, value.get()) }
    }

  fun monoEvicter(cacheStore: Cache?): CacheMono.MonoCacheEvicter<Any> =
    CacheMono.MonoCacheEvicter { key, error ->
      log.error("[${cacheStore?.name ?: cacheTag}] evicting cache $key", error)
      Mono.fromRunnable { (cacheStore ?: defaultCacheStore).evict(key) }
    }

  inline fun <reified V> fluxReader(cacheStore: Cache?): CacheFlux.FluxCacheReader<Any, V> =
    CacheFlux.FluxCacheReader<Any, V> { key ->
      log.debug("[${cacheStore?.name ?: cacheTag}] reading cache $key")
      (cacheStore ?: defaultCacheStore).get(key)
        ?.let {
          Flux.fromIterable(it.get() as List<V>)
            .materialize()
            .collectList()
        }
        ?: Mono.empty()
    }

  inline fun <reified V> fluxWriter(cacheStore: Cache?): CacheFlux.FluxCacheWriter<Any, V> =
    CacheFlux.FluxCacheWriter<Any, V> { key, value: List<Signal<V>> ->
      log.debug("[${cacheStore?.name ?: cacheTag}] writing cache $key")
      Mono.fromRunnable { (cacheStore ?: defaultCacheStore).put(
        key, value.asSequence()
        .filter { it.hasValue() }
        .map { it.get() }
        .toList()
      ) }
    }

  fun fluxEvicter(cacheStore: Cache?): CacheFlux.FluxCacheEvicter<Any> =
    CacheFlux.FluxCacheEvicter { key, error ->
      log.error("[${cacheStore?.name ?: cacheTag}] evicting cache $key", error)
      Mono.fromRunnable { (cacheStore ?: defaultCacheStore).evict(key) }
    }
}
