package eu.nk2.drm

import com.google.gson.Gson
import eu.nk2.drm.cascade.DrmSecured
import eu.nk2.drm.data.DrmResponse
import io.netty.channel.ChannelOption
import io.netty.handler.timeout.ReadTimeoutHandler
import io.netty.handler.timeout.WriteTimeoutHandler
import org.slf4j.LoggerFactory
import org.springframework.beans.factory.config.BeanPostProcessor
import org.springframework.cglib.proxy.Enhancer
import org.springframework.cglib.proxy.MethodInterceptor
import org.springframework.context.event.ContextRefreshedEvent
import org.springframework.context.event.EventListener
import org.springframework.http.client.reactive.ReactorClientHttpConnector
import org.springframework.stereotype.Service
import org.springframework.web.reactive.function.client.WebClient
import reactor.core.Disposable
import reactor.netty.http.client.HttpClient
import reactor.netty.tcp.TcpClient
import reactor.retry.Repeat
import reactor.util.retry.Retry
import java.time.Duration
import java.util.concurrent.atomic.AtomicBoolean
import javax.annotation.PreDestroy

@Service class DrmServiceBeanPostProcessor(
    private val url: String,
    private val connectTimeoutMillis: Int,
    private val retryMaxAttempts: Long,
    private val retryTimeoutMillis: Long
): BeanPostProcessor {
    private val log = LoggerFactory.getLogger(DrmServiceBeanPostProcessor::class.java)

    private val gson: Gson = Gson()
    private val webClient = WebClient.builder()
        .clientConnector(ReactorClientHttpConnector(HttpClient.from(
            TcpClient.create()
                .option(ChannelOption.CONNECT_TIMEOUT_MILLIS, connectTimeoutMillis)
                .doOnConnected { connection ->
                    connection.addHandlerLast(ReadTimeoutHandler(connectTimeoutMillis / 1000))
                        .addHandlerLast(WriteTimeoutHandler(connectTimeoutMillis / 1000))
                }
        )))
        .build()

    @Volatile private var drmIsRunning: AtomicBoolean = AtomicBoolean(true)
    @Volatile private lateinit var webClientDisposable: Disposable

    @Volatile private var drmIsUnlocked: AtomicBoolean = AtomicBoolean(true)

    @EventListener fun init(event: ContextRefreshedEvent) {
        log.info("Starting DRM with these options: url=$url, " +
            "connectTimeoutMillis=$connectTimeoutMillis, " +
            "retryMaxAttempts=$retryMaxAttempts, " +
            "retryTimeoutMillis=$retryTimeoutMillis")
        webClientDisposable = webClient.get()
            .uri(url)
            .retrieve()
            .bodyToMono(String::class.java)
            .map { gson.fromJson(it, DrmResponse::class.java) }
            .retryWhen(Retry.fixedDelay(retryMaxAttempts, Duration.ofMillis(retryTimeoutMillis)))
            .repeatWhen(Repeat.onlyIf<String> { drmIsRunning.get() }.fixedBackoff(Duration.ofMillis(retryTimeoutMillis)))
            .subscribe({
                log.debug("Received DRM info from server: $it")
                drmIsUnlocked.set(it.paid)
            }, {
                log.info("Did not receive DRM info from server: $it")
                drmIsUnlocked.set(false)
            })
    }

    @PreDestroy fun close() {
        drmIsRunning.set(false)
        webClientDisposable.dispose()
    }

    private val drmAwareBeans = hashMapOf<String, Class<*>>()

    override fun postProcessBeforeInitialization(bean: Any, beanName: String): Any? =
        bean.apply {
            if (bean.javaClass.methods.any { it.isAnnotationPresent(DrmSecured::class.java) }) {
                if(bean.javaClass.constructors.none { it.parameters.isEmpty() })
                    error("Classes that contain methods annotated with @DrmSecured must contain empty constructor, " +
                        "please look at the implementation of ${bean.javaClass.name}")
                drmAwareBeans[beanName] = this.javaClass
            }
        }

    override fun postProcessAfterInitialization(bean: Any, beanName: String): Any? =
        drmAwareBeans[beanName]?.let {
            Enhancer().apply {
                setSuperclass(it)
                setCallback(MethodInterceptor { _, method, args, _ ->
                    if(method.isAnnotationPresent(DrmSecured::class.java) && !drmIsUnlocked.get())
                        error("DRM is locked! Look at $url or contact server creator for more info")
                    method.invoke(bean, *args)
                })
            }
            .create()
        } ?: bean
}
