package eu.nk2.graphql.scalar

import graphql.language.StringValue
import graphql.schema.*
import java.time.DateTimeException
import java.time.OffsetDateTime
import java.time.ZoneOffset
import java.time.ZonedDateTime
import java.time.format.DateTimeFormatter
import java.time.format.DateTimeParseException
import java.util.*


class DateScalar: GraphQLScalarType(
  "Date",
  "An RFC-3339 compliant Full OffsetDateTime Scalar",
  object : Coercing<OffsetDateTime?, String?> {
    @Throws(CoercingSerializeException::class)
    override fun serialize(input: Any): String? {
      val offsetDateTime: OffsetDateTime = when (input) {
        is OffsetDateTime -> input
        is ZonedDateTime -> input.toOffsetDateTime()
        is Date -> input.toInstant().atOffset(ZoneOffset.UTC)
        is String -> parseOffsetDateTime(input.toString()) { CoercingSerializeException() }
        else -> throw CoercingSerializeException("Expected something we can convert to 'java.time.OffsetDateTime' but was '" + typeName(input) + "'.")
      }
      return try {
        DateTimeFormatter.ISO_OFFSET_DATE_TIME.format(offsetDateTime)
      } catch (e: DateTimeException) {
        throw CoercingSerializeException("Unable to turn TemporalAccessor into OffsetDateTime because of : '" + e.message + "'.")
      }
    }

    @Throws(CoercingParseValueException::class)
    override fun parseValue(input: Any): OffsetDateTime? =
      when (input) {
        is OffsetDateTime -> input
        is ZonedDateTime -> input.toOffsetDateTime()
        is Date -> input.toInstant().atOffset(ZoneOffset.UTC)
        is String -> parseOffsetDateTime(input.toString()) { CoercingParseValueException() }
        else -> throw CoercingParseValueException("Expected a 'String' but was '" + typeName(input) + "'.")
      }

    @Throws(CoercingParseLiteralException::class)
    override fun parseLiteral(input: Any): OffsetDateTime? {
      if (input !is StringValue) throw CoercingParseLiteralException("Expected AST type 'StringValue' but was '" + typeName(input) + "'.")
      return parseOffsetDateTime(input.value) { CoercingParseLiteralException(it) }
    }

    private fun parseOffsetDateTime(s: String, exceptionMaker: (String) -> RuntimeException): OffsetDateTime =
      try {
        OffsetDateTime.parse(s, DateTimeFormatter.ISO_OFFSET_DATE_TIME)
      } catch (e: DateTimeParseException) {
        throw exceptionMaker.invoke("Invalid RFC3339 value : '" + s + "'. because of : '" + e.message + "'")
      }
  }
) {
  companion object {
    private fun typeName(input: Any?): String =
      input?.javaClass?.simpleName ?: "null"
  }
}
