package eu.nk2.utils

import com.github.salomonbrys.kotson.*
import com.google.gson.Gson
import org.springframework.data.redis.serializer.RedisSerializer
import java.util.concurrent.ConcurrentHashMap

class GsonRedisSerializer(
    private val gson: Gson
): RedisSerializer<Any> {
    data class RedisEntry(
        val entry: Any?,
        val clazz: Class<out Any>?,
        val type: RedisEntryType
    ) {
        enum class RedisEntryType {
            NULL,
            COMMON,
            LIST,
            MAP
        }

        companion object {
            private val classCache = ConcurrentHashMap<String, Class<*>>()
            private fun <T> getClassByCache(className: String): Class<T> =
                (classCache[className] as? Class<T>) ?: (Class.forName(className) as? Class<T>)
                    ?.apply { classCache[className] = this }
                    ?: error("can't cast $className to its type")

            val serializer = jsonSerializer<RedisEntry> { (src, _, context) -> jsonObject(
                "entry" to context.serialize(src.entry),
                "class" to context.serialize(src.clazz?.name),
                "type" to src.type.name
            ) }

            val deserializer = jsonDeserializer<RedisEntry> { (src, _, context) ->
                val type = RedisEntryType.valueOf(src["type"].asString)
                val clazz = if(!src["class"].isJsonNull) getClassByCache<Any>(src["class"].asString) else null
              RedisEntry(
                entry = when (type) {
                  RedisEntryType.NULL -> null
                  RedisEntryType.COMMON -> context.deserialize(src["entry"], clazz!!)
                  RedisEntryType.LIST -> src["entry"].asJsonArray.map { context.deserialize<RedisEntry>(it) }
                  RedisEntryType.MAP -> src["entry"].asJsonObject.entrySet()
                    .asSequence()
                    .map { (key, value) -> key to context.deserialize<RedisEntry>(value) }
                    .toMap()
                },
                type = type,
                clazz = clazz
              )
            }
        }
    }

    override fun serialize(t: Any?): ByteArray? =
        t?.let {
            gson.toJson(collapseEntry(it))
                .toByteArray()
        }

    override fun deserialize(bytes: ByteArray?): Any? =
        bytes?.let {
            expandEntry(gson.fromJson(String(it), RedisEntry::class.java))
        }

    private fun collapseEntry(data: Any?): RedisEntry =
        when(data) {
            is Iterable<*> -> RedisEntry(
              entry = data.map { collapseEntry(it) },
              clazz = List::class.java,
              type = RedisEntry.RedisEntryType.LIST
            )
            is Map<*, *> -> RedisEntry(
              entry = data.map { (key, value) -> (key as String) to collapseEntry(value) }
                .toMap(),
              clazz = Map::class.java,
              type = RedisEntry.RedisEntryType.MAP
            )
            null -> RedisEntry(
              entry = null,
              clazz = null,
              type = RedisEntry.RedisEntryType.NULL
            )
            else -> RedisEntry(
              entry = data,
              clazz = data.javaClass,
              type = RedisEntry.RedisEntryType.COMMON
            )
        }

    private fun expandEntry(entry: RedisEntry): Any? =
        when(entry.type) {
            RedisEntry.RedisEntryType.NULL -> null
            RedisEntry.RedisEntryType.COMMON -> entry.entry
            RedisEntry.RedisEntryType.LIST -> (entry.entry as Iterable<RedisEntry>).map { expandEntry(it) }
            RedisEntry.RedisEntryType.MAP -> (entry.entry as Map<String, RedisEntry>).map { (key, value) -> key to expandEntry(value) }
                .toMap()
        }
}
